@@ -39,11 +39,15 @@ package gnu.javax.management;
 
 import java.lang.reflect.Array;
 import java.lang.reflect.Method;
+import java.lang.reflect.Modifier;
+import java.lang.reflect.ParameterizedType;
 import java.lang.reflect.Proxy;
 import java.lang.reflect.Type;
 
 import java.util.ArrayList;
+import java.util.Collection;
 import java.util.HashMap;
+import java.util.Iterator;
 import java.util.List;
 import java.util.Map;
 import java.util.Set;
@@ -92,7 +96,7 @@ public final class Translator
     Type[] gtypes = method.getGenericParameterTypes();
     Object[] otypes = new Object[jtypes.length];
     for (int a = 0; a < jtypes.length; ++a)
-      otypes[a] = fromJava(jtypes[a], (Class<?>) gtypes[a]);
+      otypes[a] = fromJava(jtypes[a], gtypes[a]);
     return otypes;
   }
 
@@ -131,30 +135,31 @@ public final class Translator
     if (jtype instanceof List || jtype instanceof Set ||
 	jtype instanceof SortedSet)
       {
-	String elemType = tName.substring(tName.indexOf("<") + 1,
-					  tName.indexOf(">")).trim();
 	if (jtype instanceof SortedSet)
 	  {
-	    Class<?> elemClass = Class.forName(elemType);
+	    ParameterizedType ptype = (ParameterizedType) type;
+	    Class<?> elemClass = (Class<?>) ptype.getActualTypeArguments()[0];
 	    if (!Comparable.class.isAssignableFrom(elemClass))
 	      throw new IllegalArgumentException(jtype + " has a " +
 						 "non-comparable element " +
 						 "type, " + elemClass);
-	    if (((SortedSet) jtype).comparator() != null)
+	    if (((SortedSet<?>) jtype).comparator() != null)
 	      throw new IllegalArgumentException(jtype + " does not " +
 						 "use natural ordering.");
 	  }
-	List elems = (List) jtype;
-	Object[] celems = new Object[elems.size()];
-	for (int a = 0; a < elems.size(); ++a)
+	Collection<?> elems = (Collection<?>) jtype;
+	int numElems = elems.size();
+	Object[] celems = new Object[numElems];
+	Iterator<?> i = elems.iterator();
+	for (int a = 0; a < numElems; ++a)
 	  {
-	    Object elem = elems.get(a);
+	    Object elem = i.next();
 	    celems[a] = fromJava(elem, elem.getClass());
 	  }
 	return makeArraySpecific(celems);
       }
     if (jtype instanceof Enum)
-      return ((Enum) jtype).name();
+      return ((Enum<?>) jtype).name();
     if (jtype instanceof Map || jtype instanceof SortedMap)
       {
 	int lparam = tName.indexOf("<");
@@ -172,13 +177,13 @@ public final class Translator
 	      throw new IllegalArgumentException(jtype + " has a " +
 						 "non-comparable element " +
 						 "type, " + keyClass);
-	    if (((SortedMap) jtype).comparator() != null)
+	    if (((SortedMap<?,?>) jtype).comparator() != null)
 	      throw new IllegalArgumentException(jtype + " does not " +
 						 "use natural ordering.");
 	    typeName = "java.util.SortedMap" + tName.substring(lparam);
 	  }
-	OpenType k = translate(key).getOpenType();
-	OpenType v = translate(value).getOpenType(); 
+	OpenType<?> k = translate(key).getOpenType();
+	OpenType<?> v = translate(value).getOpenType(); 
 	CompositeType rowType = new CompositeType(typeName, typeName,
 						  new String[] { "key", "value" },
 						  new String[] { "Map key", "Map value"},
@@ -186,7 +191,7 @@ public final class Translator
 	TabularType tabType = new TabularType(typeName, typeName, rowType,
 					      new String[]{"key"});
 	TabularData data = new TabularDataSupport(tabType);
-	for (Map.Entry entry : (Set<Map.Entry>) ((Map) jtype).entrySet())
+	for (Map.Entry<?,?> entry : ((Map<?,?>) jtype).entrySet())
 	  {
 	    try 
 	      {
@@ -253,15 +258,15 @@ public final class Translator
     if (returnType.isEnum())
       {
 	String ename = (String) otype;
-	Enum[] constants = (Enum[]) returnType.getEnumConstants();
-	for (Enum c : constants)
+	Enum<?>[] constants = (Enum[]) returnType.getEnumConstants();
+	for (Enum<?> c : constants)
 	  if (c.name().equals(ename))
 	    return c;
       }
     if (List.class.isAssignableFrom(returnType))
       {
 	Object[] elems = (Object[]) otype;
-	List l = new ArrayList(elems.length);
+	List<Object> l = new ArrayList<Object>(elems.length);
 	for (Object elem : elems)
 	  l.add(elem);
 	return l;
@@ -269,7 +274,7 @@ public final class Translator
     if (Map.class.isAssignableFrom(returnType))
       {
 	TabularData data = (TabularData) otype;
-	Map m = new HashMap(data.size());
+	Map<Object,Object> m = new HashMap<Object,Object>(data.size());
 	for (Object val : data.values())
 	  {
 	    CompositeData vals = (CompositeData) val;
@@ -396,8 +401,8 @@ public final class Translator
 	int comma = type.indexOf(",", lparam);
 	int rparam = type.indexOf(">", comma);
 	String key = type.substring(lparam + 1, comma).trim();
-	OpenType k = translate(key).getOpenType();
-	OpenType v = translate(type.substring(comma + 1, rparam).trim()).getOpenType(); 
+	OpenType<?> k = translate(key).getOpenType();
+	OpenType<?> v = translate(type.substring(comma + 1, rparam).trim()).getOpenType(); 
  	CompositeType ctype = new CompositeType(Map.class.getName(), Map.class.getName(),
 						new String[] { "key", "value" },
 						new String[] { "Map key", "Map value"},
@@ -412,13 +417,13 @@ public final class Translator
       {
 	int lparam = type.indexOf("<");
 	int rparam = type.indexOf(">");
-       	OpenType e = translate(type.substring(lparam + 1, rparam).trim()).getOpenType();
+       	OpenType<?> e = translate(type.substring(lparam + 1, rparam).trim()).getOpenType();
 	return new OpenMBeanParameterInfoSupport("TransParam",
 						 "Translated parameter",
-						 new ArrayType(1, e)
+						 new ArrayType<OpenType<?>>(1, e)
 						 );
       }	
-    Class c;
+    Class<?> c;
     try
       {
 	c = Class.forName(type);
@@ -440,45 +445,46 @@ public final class Translator
 						 SimpleType.STRING,
 						 null, names);
       }
-    try
-      {
-	c.getMethod("from", new Class[] { CompositeData.class });
-	Method[] methods = c.getDeclaredMethods();
-	List<String> names = new ArrayList<String>();
-	List<OpenType> types = new ArrayList<OpenType>();
-	for (int a = 0; a < methods.length; ++a)
-	  {
-	    String name = methods[a].getName();
-	    if (name.startsWith("get"))
-	      {
-		names.add(name.substring(3));
-		types.add(getTypeFromClass(methods[a].getReturnType()));
-	      }
-	  }
-	String[] fields = names.toArray(new String[names.size()]);
-	CompositeType ctype = new CompositeType(c.getName(), c.getName(),
-						fields, fields,
-						types.toArray(new OpenType[types.size()]));
-	return new OpenMBeanParameterInfoSupport("TransParam",
-						 "Translated parameter",
-						 ctype);
-      }
-    catch (NoSuchMethodException e)
-      {
-	/* Ignored; we expect this if this isn't a from(CompositeData) class */
-      }
     if (c.isArray())
       {
 	int depth;
 	for (depth = 0; c.getName().charAt(depth) == '['; ++depth)
           ;
-	OpenType ot = getTypeFromClass(c.getComponentType());
+	OpenType<?> ot = getTypeFromClass(c.getComponentType());
 	return new OpenMBeanParameterInfoSupport("TransParam",
 						 "Translated parameter",
-						 new ArrayType(depth, ot)
+						 new ArrayType<OpenType<?>>(depth, ot)
 						 );
       }
-    throw new InternalError("The type used does not have an open type translation.");
+    Method[] methods = c.getDeclaredMethods();
+    List<String> names = new ArrayList<String>();
+    List<OpenType<?>> types = new ArrayList<OpenType<?>>();
+    for (int a = 0; a < methods.length; ++a)
+      {
+	String name = methods[a].getName();
+	if (Modifier.isPublic(methods[a].getModifiers()))
+	  {
+	    if (name.startsWith("get"))
+	      {
+		names.add(name.substring(3));
+		types.add(getTypeFromClass(methods[a].getReturnType()));
+	      }
+	    else if (name.startsWith("is"))
+	      {
+		names.add(name.substring(2));
+		types.add(getTypeFromClass(methods[a].getReturnType()));
+	      }
+	  }
+      }
+    if (names.isEmpty())
+      throw new OpenDataException("The type used does not have an open type translation.");
+    String[] fields = names.toArray(new String[names.size()]);
+    CompositeType ctype = new CompositeType(c.getName(), c.getName(),
+					    fields, fields,
+					    types.toArray(new OpenType[types.size()]));
+    return new OpenMBeanParameterInfoSupport("TransParam",
+					     "Translated parameter",
+					     ctype);
   }
 
   /**
@@ -489,7 +495,7 @@ public final class Translator
    * @return the appropriate instance.
    * @throws OpenDataException if the type is not open.
    */
-  private static final OpenType getTypeFromClass(Class c)
+  private static final OpenType<?> getTypeFromClass(Class<?> c)
     throws OpenDataException
   {
     return Translator.translate(c.getName()).getOpenType();
