@@ -38,16 +38,26 @@ exception statement from your version. */
 
 package javax.swing;
 
+import gnu.classpath.SystemProperties;
+import gnu.java.awt.LowPriorityEvent;
+
+import java.applet.Applet;
 import java.awt.Component;
 import java.awt.Dimension;
+import java.awt.EventQueue;
+import java.awt.Graphics;
 import java.awt.Image;
 import java.awt.Rectangle;
+import java.awt.Toolkit;
+import java.awt.Window;
+import java.awt.event.InvocationEvent;
 import java.awt.image.VolatileImage;
 import java.util.ArrayList;
-import java.util.Collections;
-import java.util.Comparator;
 import java.util.HashMap;
+import java.util.HashSet;
 import java.util.Iterator;
+import java.util.Set;
+import java.util.WeakHashMap;
 
 /**
  * <p>The repaint manager holds a set of dirty regions, invalid components,
@@ -57,19 +67,65 @@ import java.util.Iterator;
  * double buffer surface is used by root components to paint
  * themselves.</p>
  *
- * <p>In general, painting is very confusing in swing. see <a
+ * <p>See <a
  * href="http://java.sun.com/products/jfc/tsc/articles/painting/index.html">this
  * document</a> for more details.</p>
+ * document</a> for more details.</p>
  *
+ * @author Roman Kennke (kennke@aicas.com)
  * @author Graydon Hoare (graydon@redhat.com)
+ * @author Audrius Meskauskas (audriusa@bioinformatics.org)
  */
 public class RepaintManager
 {
+  /**
+   * An InvocationEvent subclass that implements LowPriorityEvent. This is used
+   * to defer the execution of RepaintManager requests as long as possible on
+   * the event queue. This way we make sure that all available input is
+   * processed before getting active with the RepaintManager. This allows
+   * for better optimization (more validate and repaint requests can be
+   * coalesced) and thus has a positive effect on performance for GUI
+   * applications under heavy load.
+   */
+  private static class RepaintWorkerEvent
+    extends InvocationEvent
+    implements LowPriorityEvent
+  {
+
+    /**
+     * Creates a new RepaintManager event.
+     *
+     * @param source the source
+     * @param runnable the runnable to execute
+     */
+    public RepaintWorkerEvent(Object source, Runnable runnable,
+                              Object notifier, boolean catchEx)
+    {
+      super(source, runnable, notifier, catchEx);
+    }
+
+    /**
+     * An application that I met implements its own event dispatching and
+     * calls dispatch() via reflection, and only checks declared methods,
+     * that is, it expects this method to be in the event's class, not
+     * in a superclass. So I put this in here... sigh.
+     */
+    public void dispatch()
+    {
+      super.dispatch();
+    }
+  }
+  
   /**
    * The current repaint managers, indexed by their ThreadGroups.
    */
-  static HashMap currentRepaintManagers;
-  
+  static WeakHashMap currentRepaintManagers;
+
+  /**
+   * A rectangle object to be reused in damaged regions calculation.
+   */
+  private static Rectangle rectCache = new Rectangle();
+
   /**
    * <p>A helper class which is placed into the system event queue at
    * various times in order to facilitate repainting and layout. There is
@@ -83,7 +139,7 @@ public class RepaintManager
    * swing paint thread, which revalidates all invalid components and
    * repaints any damage in the swing scene.</p>
    */
-  protected class RepaintWorker
+  private class RepaintWorker
     implements Runnable
   {
 
@@ -106,70 +162,20 @@ public class RepaintManager
 
     public void run()
     {
-      ThreadGroup threadGroup = Thread.currentThread().getThreadGroup();
-      RepaintManager rm =
-        (RepaintManager) currentRepaintManagers.get(threadGroup);
-      setLive(false);
-      rm.validateInvalidComponents();
-      rm.paintDirtyRegions();
-    }
-
-  }
-
-  /**
-   * Compares two components using their depths in the component hierarchy.
-   * A component with a lesser depth (higher level components) are sorted
-   * before components with a deeper depth (low level components). This is used
-   * to order paint requests, so that the higher level components are painted
-   * before the low level components get painted.
-   *
-   * @author Roman Kennke (kennke@aicas.com)
-   */
-  private class ComponentComparator implements Comparator
-  {
-
-    /**
-     * Compares two components.
-     *
-     * @param o1 the first component
-     * @param o2 the second component
-     *
-     * @return a negative integer, if <code>o1</code> is higher in the
-     *         hierarchy than <code>o2</code>, zero, if both are at the same
-     *         level and a positive integer, if <code>o1</code> is deeper in
-     *         the hierarchy than <code>o2</code> 
-     */
-    public int compare(Object o1, Object o2)
-    {
-      if (o1 instanceof JComponent && o2 instanceof JComponent)
+      try
         {
-          JComponent c1 = (JComponent) o1;
-          JComponent c2 = (JComponent) o2;
-          return getDepth(c1) - getDepth(c2);
+          ThreadGroup threadGroup = Thread.currentThread().getThreadGroup();
+          RepaintManager rm =
+            (RepaintManager) currentRepaintManagers.get(threadGroup);
+          rm.validateInvalidComponents();
+          rm.paintDirtyRegions();
+        }
+      finally
+        {
+          setLive(false);
         }
-      else
-        throw new ClassCastException("This comparator can only be used with "
-                                     + "JComponents");
     }
 
-    /**
-     * Computes the depth for a given JComponent.
-     *
-     * @param c the component to compute the depth for
-     *
-     * @return the depth of the component
-     */
-    private int getDepth(JComponent c)
-    {
-      Component comp = c;
-      int depth = 0;
-      while (comp != null)
-        {
-          comp = comp.getParent();
-          depth++;
-        }
-      return depth;
-    }
   }
 
   /** 
@@ -178,26 +184,22 @@ public class RepaintManager
    * to exactly one rectangle.  When more regions are marked as dirty on a
    * component, they are union'ed with the existing rectangle.
    *
+   * This is package private to avoid a synthetic accessor method in inner
+   * class.
+   *
    * @see #addDirtyRegion
    * @see #getDirtyRegion
    * @see #isCompletelyDirty
    * @see #markCompletelyClean
    * @see #markCompletelyDirty
    */
-  HashMap dirtyComponents;
-
-  HashMap workDirtyComponents;
+  private HashMap dirtyComponents;
 
   /**
-   * Stores the order in which the components get repainted.
+   * The dirtyComponents which is used in paintDiryRegions to avoid unnecessary
+   * locking.
    */
-  ArrayList repaintOrder;
-  ArrayList workRepaintOrder;
-
-  /**
-   * The comparator used for ordered inserting into the repaintOrder list. 
-   */
-  Comparator comparator;
+  private HashMap dirtyComponentsWork;
 
   /**
    * A single, shared instance of the helper class. Any methods which mark
@@ -208,7 +210,7 @@ public class RepaintManager
    * @see #addDirtyRegion
    * @see #addInvalidComponent
    */
-  RepaintWorker repaintWorker;
+  private RepaintWorker repaintWorker;
 
   /** 
    * The set of components which need revalidation, in the "layout" sense.
@@ -220,8 +222,7 @@ public class RepaintManager
    * @see #removeInvalidComponent
    * @see #validateInvalidComponents
    */
-  ArrayList invalidComponents;
-  ArrayList workInvalidComponents;
+  private ArrayList invalidComponents;
 
   /** 
    * Whether or not double buffering is enabled on this repaint
@@ -231,17 +232,14 @@ public class RepaintManager
    * @see #isDoubleBufferingEnabled
    * @see #setDoubleBufferingEnabled
    */
-  boolean doubleBufferingEnabled;
+  private boolean doubleBufferingEnabled;
 
-  /** 
-   * The current offscreen buffer. This is reused for all requests for
-   * offscreen drawing buffers. It grows as necessary, up to {@link
-   * #doubleBufferMaximumSize}, but there is only one shared instance.
-   *
-   * @see #getOffscreenBuffer
-   * @see #doubleBufferMaximumSize
+  /**
+   * The offscreen buffers. This map holds one offscreen buffer per
+   * Window/Applet and releases them as soon as the Window/Applet gets garbage
+   * collected.
    */
-  Image doubleBuffer;
+  private WeakHashMap offscreenBuffers;
 
   /**
    * The maximum width and height to allocate as a double buffer. Requests
@@ -251,7 +249,7 @@ public class RepaintManager
    * @see #getDoubleBufferMaximumSize
    * @see #setDoubleBufferMaximumSize
    */
-  Dimension doubleBufferMaximumSize;
+  private Dimension doubleBufferMaximumSize;
 
 
   /**
@@ -260,14 +258,14 @@ public class RepaintManager
   public RepaintManager()
   {
     dirtyComponents = new HashMap();
-    workDirtyComponents = new HashMap();
-    repaintOrder = new ArrayList();
-    workRepaintOrder = new ArrayList();
+    dirtyComponentsWork = new HashMap();
     invalidComponents = new ArrayList();
-    workInvalidComponents = new ArrayList();
     repaintWorker = new RepaintWorker();
     doubleBufferMaximumSize = new Dimension(2000,2000);
-    doubleBufferingEnabled = true;
+    doubleBufferingEnabled =
+      SystemProperties.getProperty("gnu.swing.doublebuffering", "true")
+                      .equals("true");
+    offscreenBuffers = new WeakHashMap();
   }
 
   /**
@@ -286,7 +284,7 @@ public class RepaintManager
   public static RepaintManager currentManager(Component component)
   {
     if (currentRepaintManagers == null)
-      currentRepaintManagers = new HashMap();
+      currentRepaintManagers = new WeakHashMap();
     ThreadGroup threadGroup = Thread.currentThread().getThreadGroup();
     RepaintManager currentManager =
       (RepaintManager) currentRepaintManagers.get(threadGroup);
@@ -330,7 +328,7 @@ public class RepaintManager
   public static void setCurrentManager(RepaintManager manager)
   {
     if (currentRepaintManagers == null)
-      currentRepaintManagers = new HashMap();
+      currentRepaintManagers = new WeakHashMap();
 
     ThreadGroup threadGroup = Thread.currentThread().getThreadGroup();
     currentRepaintManagers.put(threadGroup, manager);
@@ -345,29 +343,51 @@ public class RepaintManager
    *
    * @see #removeInvalidComponent
    */
-  public synchronized void addInvalidComponent(JComponent component)
+  public void addInvalidComponent(JComponent component)
   {
-    Component ancestor = component.getParent();
+    Component validateRoot = null;
+    Component c = component;
+    while (c != null)
+      {
+        // Special cases we don't bother validating are when the invalidated
+        // component (or any of it's ancestors) is inside a CellRendererPane
+        // or if it doesn't have a peer yet (== not displayable).
+        if (c instanceof CellRendererPane || ! c.isDisplayable())
+          return;
+        if (c instanceof JComponent && ((JComponent) c).isValidateRoot())
+          {
+            validateRoot = c;
+            break;
+          }
 
-    while (ancestor != null
-           && (! (ancestor instanceof JComponent)
-               || ! ((JComponent) ancestor).isValidateRoot() ))
-      ancestor = ancestor.getParent();
+        c = c.getParent();
+      }
 
-    if (ancestor != null
-        && ancestor instanceof JComponent
-        && ((JComponent) ancestor).isValidateRoot())
-      component = (JComponent) ancestor;
-
-    if (invalidComponents.contains(component))
+    // If we didn't find a validate root, then we don't validate.
+    if (validateRoot == null)
       return;
 
-    invalidComponents.add(component);
-    
+    // Make sure the validate root and all of it's ancestors are visible.
+    c = validateRoot;
+    while (c != null)
+      {
+        if (! c.isVisible() || ! c.isDisplayable())
+          return;
+        c = c.getParent();
+      }
+
+    if (invalidComponents.contains(validateRoot))
+      return;
+
+    //synchronized (invalidComponents)
+    //  {
+        invalidComponents.add(validateRoot);
+    //  }
+
     if (! repaintWorker.isLive())
       {
         repaintWorker.setLive(true);
-        SwingUtilities.invokeLater(repaintWorker);
+        invokeLater(repaintWorker);
       }
   }
 
@@ -378,9 +398,12 @@ public class RepaintManager
    *
    * @see #addInvalidComponent
    */
-  public synchronized void removeInvalidComponent(JComponent component)
+  public void removeInvalidComponent(JComponent component)
   {
-    invalidComponents.remove(component);
+    synchronized (invalidComponents)
+      {
+        invalidComponents.remove(component);
+      }
   }
 
   /**
@@ -401,38 +424,37 @@ public class RepaintManager
    * @see #markCompletelyClean
    * @see #markCompletelyDirty
    */
-  public synchronized void addDirtyRegion(JComponent component, int x, int y,
-                                          int w, int h)
+  public void addDirtyRegion(JComponent component, int x, int y,
+                             int w, int h)
   {
-    if (w == 0 || h == 0 || !component.isShowing())
+    if (w <= 0 || h <= 0 || !component.isShowing())
       return;
-    Rectangle r = new Rectangle(x, y, w, h);
-    if (dirtyComponents.containsKey(component))
-      r = r.union((Rectangle)dirtyComponents.get(component));
-    else
-      insertInRepaintOrder(component);
-    dirtyComponents.put(component, r);
-    if (! repaintWorker.isLive())
-      {
-        repaintWorker.setLive(true);
-        SwingUtilities.invokeLater(repaintWorker);
-      }
-  }
+    component.computeVisibleRect(rectCache);
+    SwingUtilities.computeIntersection(x, y, w, h, rectCache);
 
-  /**
-   * Inserts a component into the repaintOrder list in an ordered fashion,
-   * using a binary search.
-   *
-   * @param c the component to be inserted
-   */
-  private void insertInRepaintOrder(JComponent c)
-  {
-    if (comparator == null)
-      comparator = new ComponentComparator();
-    int insertIndex = Collections.binarySearch(repaintOrder, c, comparator);
-    if (insertIndex < 0)
-      insertIndex = -(insertIndex + 1);
-    repaintOrder.add(insertIndex, c);
+    if (! rectCache.isEmpty())
+      {
+        synchronized (dirtyComponents)
+          {
+            Rectangle dirtyRect = (Rectangle)dirtyComponents.get(component);
+            if (dirtyRect != null)
+              {
+                SwingUtilities.computeUnion(rectCache.x, rectCache.y,
+                                            rectCache.width, rectCache.height,
+                                            dirtyRect);
+              }
+            else
+              {
+                dirtyComponents.put(component, rectCache.getBounds());
+              }
+          }
+
+        if (! repaintWorker.isLive())
+          {
+            repaintWorker.setLive(true);
+            invokeLater(repaintWorker);
+          }
+      }
   }
 
   /**
@@ -470,9 +492,7 @@ public class RepaintManager
    */
   public void markCompletelyDirty(JComponent component)
   {
-    Rectangle r = component.getBounds();
-    addDirtyRegion(component, r.x, r.y, r.width, r.height);
-    component.isCompletelyDirty = true;
+    addDirtyRegion(component, 0, 0, Integer.MAX_VALUE, Integer.MAX_VALUE);
   }
 
   /**
@@ -488,11 +508,10 @@ public class RepaintManager
    */
   public void markCompletelyClean(JComponent component)
   {
-    synchronized (this)
+    synchronized (dirtyComponents)
       {
         dirtyComponents.remove(component);
       }
-    component.isCompletelyDirty = false;
   }
 
   /**
@@ -511,9 +530,11 @@ public class RepaintManager
    */
   public boolean isCompletelyDirty(JComponent component)
   {
-    if (! dirtyComponents.containsKey(component))
-      return false;
-    return component.isCompletelyDirty;
+    boolean dirty = false;
+    Rectangle r = getDirtyRegion(component);
+    if(r.width == Integer.MAX_VALUE && r.height == Integer.MAX_VALUE)
+      dirty = true;
+    return dirty;
   }
 
   /**
@@ -522,56 +543,150 @@ public class RepaintManager
    */
   public void validateInvalidComponents()
   {
-    // In order to keep the blocking of application threads minimal, we switch
-    // the invalidComponents field with the workInvalidComponents field and
-    // work with the workInvalidComponents field.
-    synchronized(this)
-    {
-      ArrayList swap = invalidComponents;
-      invalidComponents = workInvalidComponents;
-      workInvalidComponents = swap;
-    }
-    for (Iterator i = workInvalidComponents.iterator(); i.hasNext(); )
+    // We don't use an iterator here because that would fail when there are
+    // components invalidated during the validation of others, which happens
+    // quite frequently. Instead we synchronize the access a little more.
+    while (invalidComponents.size() > 0)
       {
-        JComponent comp = (JComponent) i.next();
+        Component comp;
+        synchronized (invalidComponents)
+          {
+            comp = (Component) invalidComponents.remove(0);
+          }
+        // Validate the validate component.
         if (! (comp.isVisible() && comp.isShowing()))
           continue;
         comp.validate();
       }
-    workInvalidComponents.clear();
   }
 
   /**
-   * Repaint all regions of all components which have been marked dirty in
-   * the {@link #dirtyComponents} table.
+   * Repaint all regions of all components which have been marked dirty in the
+   * {@link #dirtyComponents} table.
    */
-  public synchronized void paintDirtyRegions()
+  public void paintDirtyRegions()
   {
-    // In order to keep the blocking of application threads minimal, we switch
-    // the dirtyComponents field with the workdirtyComponents field and the
-    // repaintOrder field with the workRepaintOrder field and work with the
-    // work* fields.
-    synchronized(this)
-    {
-      ArrayList swap = workRepaintOrder;
-      workRepaintOrder = repaintOrder;
-      repaintOrder = swap;
-      HashMap swap2 = workDirtyComponents;
-      workDirtyComponents = dirtyComponents;
-      dirtyComponents = swap2;
-    }
-    for (Iterator i = workRepaintOrder.iterator(); i.hasNext();)
+    // Short circuit if there is nothing to paint.
+    if (dirtyComponents.size() == 0)
+      return;
+
+    // Swap dirtyRegions with dirtyRegionsWork to avoid locking.
+    synchronized (dirtyComponents)
+      {
+        HashMap swap = dirtyComponents;
+        dirtyComponents = dirtyComponentsWork;
+        dirtyComponentsWork = swap;
+      }
+
+    // Compile a set of repaint roots.
+    HashSet repaintRoots = new HashSet();
+    Set components = dirtyComponentsWork.keySet();
+    for (Iterator i = components.iterator(); i.hasNext();)
+      {
+        JComponent dirty = (JComponent) i.next();
+        compileRepaintRoots(dirtyComponentsWork, dirty, repaintRoots);
+      }
+
+    for (Iterator i = repaintRoots.iterator(); i.hasNext();)
       {
         JComponent comp = (JComponent) i.next();
-        // If a component is marked completely clean in the meantime, then skip
-        // it.
-        Rectangle damaged = (Rectangle) workDirtyComponents.get(comp);
+        Rectangle damaged = (Rectangle) dirtyComponentsWork.remove(comp);
         if (damaged == null || damaged.isEmpty())
           continue;
         comp.paintImmediately(damaged);
       }
-    workRepaintOrder.clear();
-    workDirtyComponents.clear();
+    dirtyComponentsWork.clear();
+  }
+
+  /**
+   * Compiles a list of components that really get repainted. This is called
+   * once for each component in the dirtyRegions HashMap, each time with
+   * another <code>dirty</code> parameter. This searches up the component
+   * hierarchy of <code>dirty</code> to find the highest parent that is also
+   * marked dirty and merges the dirty regions.
+   *
+   * @param dirtyRegions the dirty regions 
+   * @param dirty the component for which to find the repaint root
+   * @param roots the list to which new repaint roots get appended
+   */
+  private void compileRepaintRoots(HashMap dirtyRegions, JComponent dirty,
+                                   HashSet roots)
+  {
+    Component current = dirty;
+    Component root = dirty;
+
+    // This will contain the dirty region in the root coordinate system,
+    // possibly clipped by ancestor's bounds.
+    Rectangle originalDirtyRect = (Rectangle) dirtyRegions.get(dirty); 
+    rectCache.setBounds(originalDirtyRect);
+
+    // The bounds of the current component.
+    int x = dirty.getX();
+    int y = dirty.getY();
+    int w = dirty.getWidth();
+    int h = dirty.getHeight();
+
+    // Do nothing if dirty region is clipped away by the component's bounds.
+    rectCache = SwingUtilities.computeIntersection(0, 0, w, h, rectCache);
+    if (rectCache.isEmpty())
+      return;
+
+    // The cumulated offsets. 
+    int dx = 0;
+    int dy = 0;
+    // The actual offset for the found root.
+    int rootDx = 0;
+    int rootDy = 0;
+
+    // Search the highest component that is also marked dirty.
+    Component parent;
+    while (true)
+      {
+        parent = current.getParent();
+        if (parent == null || !(parent instanceof JComponent))
+          break;
+
+        current = parent;
+        // Update the offset.
+        dx += x;
+        dy += y;
+        rectCache.x += x;
+        rectCache.y += y;
+        
+        x = current.getX();
+        y = current.getY();
+        w = current.getWidth();
+        h = current.getHeight();
+        rectCache = SwingUtilities.computeIntersection(0, 0, w, h, rectCache);
+
+        // Don't paint if the dirty regions is clipped away by any of
+        // its ancestors.
+        if (rectCache.isEmpty())
+          return;
+
+        // We can skip to the next up when this parent is not dirty.
+        if (dirtyRegions.containsKey(parent))
+          {
+            root = current;
+            rootDx = dx;
+            rootDy = dy;
+          }
+      }
+
+    // Merge the rectangles of the root and the requested component if
+    // the are different.
+    if (root != dirty)
+      {
+        rectCache.x += rootDx - dx;
+        rectCache.y += rootDy - dy;
+        Rectangle dirtyRect = (Rectangle) dirtyRegions.get(root);
+        SwingUtilities.computeUnion(rectCache.x, rectCache.y, rectCache.width,
+                                    rectCache.height, dirtyRect);
+      }
+
+    // Adds the root to the roots set.
+    if (! roots.contains(root))
+      roots.add(root);
   }
 
   /**
@@ -584,21 +699,59 @@ public class RepaintManager
    * @param proposedHeight The proposed height of the offscreen buffer
    *
    * @return A shared offscreen buffer for painting
-   *
-   * @see #doubleBuffer
    */
   public Image getOffscreenBuffer(Component component, int proposedWidth,
                                   int proposedHeight)
   {
-    if (doubleBuffer == null 
-        || (((doubleBuffer.getWidth(null) < proposedWidth) 
-             || (doubleBuffer.getHeight(null) < proposedHeight))
-            && (proposedWidth < doubleBufferMaximumSize.width)
-            && (proposedHeight < doubleBufferMaximumSize.height)))
+    Component root = SwingUtilities.getWindowAncestor(component);
+    Image buffer = (Image) offscreenBuffers.get(root);
+    if (buffer == null 
+        || buffer.getWidth(null) < proposedWidth 
+        || buffer.getHeight(null) < proposedHeight)
       {
-        doubleBuffer = component.createImage(proposedWidth, proposedHeight);
+        int width = Math.max(proposedWidth, root.getWidth());
+        width = Math.min(doubleBufferMaximumSize.width, width);
+        int height = Math.max(proposedHeight, root.getHeight());
+        height = Math.min(doubleBufferMaximumSize.height, height);
+        buffer = component.createImage(width, height);
+        offscreenBuffers.put(root, buffer);
+      }
+    return buffer;
+  }
+
+  /**
+   * Blits the back buffer of the specified root component to the screen.
+   * This is package private because it must get called by JComponent.
+   *
+   * @param comp the component to be painted
+   * @param x the area to paint on screen, in comp coordinates
+   * @param y the area to paint on screen, in comp coordinates
+   * @param w the area to paint on screen, in comp coordinates
+   * @param h the area to paint on screen, in comp coordinates
+   */
+  void commitBuffer(Component comp, int x, int y, int w, int h)
+  {
+    Component root = comp;
+    while (root != null
+	   && ! (root instanceof Window || root instanceof Applet))
+      {
+	x += root.getX();
+	y += root.getY();
+	root = root.getParent();
+      }
+
+    if (root != null)
+      {
+        Graphics g = root.getGraphics();
+        Image buffer = (Image) offscreenBuffers.get(root);
+        if (buffer != null)
+          {
+            // Make sure we have a sane clip at this point.
+            g.clipRect(x, y, w, h);
+            g.drawImage(buffer, 0, 0, root);
+            g.dispose();
+          }
       }
-    return doubleBuffer;
   }
 
   /**
@@ -619,10 +772,22 @@ public class RepaintManager
   public Image getVolatileOffscreenBuffer(Component comp, int proposedWidth,
                                           int proposedHeight)
   {
-    int maxWidth = doubleBufferMaximumSize.width;
-    int maxHeight = doubleBufferMaximumSize.height;
-    return comp.createVolatileImage(Math.min(maxWidth, proposedWidth),
-                                    Math.min(maxHeight, proposedHeight));
+    Component root = SwingUtilities.getWindowAncestor(comp);
+    Image buffer = (Image) offscreenBuffers.get(root);
+    if (buffer == null 
+        || buffer.getWidth(null) < proposedWidth 
+        || buffer.getHeight(null) < proposedHeight
+        || !(buffer instanceof VolatileImage))
+      {
+        int width = Math.max(proposedWidth, root.getWidth());
+        width = Math.min(doubleBufferMaximumSize.width, width);
+        int height = Math.max(proposedHeight, root.getHeight());
+        height = Math.min(doubleBufferMaximumSize.height, height);
+        buffer = root.createVolatileImage(width, height);
+        if (buffer != null)
+          offscreenBuffers.put(root, buffer);
+      }
+    return buffer;
   }
   
 
@@ -678,4 +843,18 @@ public class RepaintManager
   {
     return "RepaintManager";
   }
+
+  /**
+   * Sends an RepaintManagerEvent to the event queue with the specified
+   * runnable. This is similar to SwingUtilities.invokeLater(), only that the
+   * event is a low priority event in order to defer the execution a little
+   * more.
+   */
+  private void invokeLater(Runnable runnable)
+  {
+    Toolkit tk = Toolkit.getDefaultToolkit();
+    EventQueue evQueue = tk.getSystemEventQueue();
+    InvocationEvent ev = new RepaintWorkerEvent(evQueue, runnable, null, false);
+    evQueue.postEvent(ev);
+  }
 }
